#include "Thing.h"

#include "Messages/PoseMsg.h"
#include "Participant.h"
#include "Participants/IsolatedParticipant.h"

#include <string.h>

#if defined(ARDUINO)
#include "Arduino.h"
#else
#include <algorithm>
#include <chrono>
#include <iostream>
#include <list>
#endif

namespace RoboidControl {

Thing::Thing(int thingType)
    : Thing(IsolatedParticipant::Isolated(), thingType) {}

Thing::Thing(Participant* owner, Type thingType, unsigned char thingId)
    : Thing(owner, (unsigned char)thingType, thingId) {}

Thing::Thing(Participant* owner, int thingType, unsigned char thingId) {
  this->owner = owner;
  this->id = thingId;
  this->type = thingType;
  this->networkId = 0;

  this->position = Spherical::zero;
  this->positionUpdated = true;
  this->orientation = SwingTwist::identity;
  this->orientationUpdated = true;
  this->hierarchyChanged = true;

  this->linearVelocity = Spherical::zero;
  this->angularVelocity = Spherical::zero;

  // std::cout << "add thing [" << (int)this->id << "] to owner "
  //           << this->owner->ipAddress << ":" << this->owner->port <<
  //           std::endl;
  this->owner->Add(this, true);
}

Thing::Thing(Thing* parent, int thingType, unsigned char thingId)
    : Thing(parent->owner, thingType, thingId) {
  this->SetParent(parent);
}

void Thing::SetName(const char* name) {
  this->name = name;
  this->nameChanged = true;
}

const char* Thing::GetName() const {
  return this->name;
}

Thing* Thing::FindThing(const char* name) {
  for (unsigned char childIx = 0; childIx < this->childCount; childIx++) {
    Thing* child = this->children[childIx];
    if (child == nullptr || child->name == nullptr)
      continue;

    if (strcmp(child->name, name) == 0)
      return child;

    Thing* foundChild = child->FindThing(name);
    if (foundChild != nullptr)
      return foundChild;
  }
  return nullptr;
}

void Thing::SetParent(Thing* parent) {
  if (parent == nullptr) {
    Thing* parentThing = this->parent;
    if (parentThing != nullptr)
      parentThing->RemoveChild(this);
    this->parent = nullptr;
  } else
    parent->AddChild(this);
  this->hierarchyChanged = true;
}

void Thing::SetParent(Thing* root, const char* name) {
  Thing* thing = root->FindThing(name);
  if (thing != nullptr)
    this->SetParent(thing);
}

Thing* Thing::GetParent() {
  return this->parent;
}

void Thing::AddChild(Thing* child) {
  unsigned char newChildCount = this->childCount + 1;
  Thing** newChildren = new Thing*[newChildCount];

  for (unsigned char childIx = 0; childIx < this->childCount; childIx++) {
    newChildren[childIx] = this->children[childIx];
    if (this->children[childIx] == child) {
      // child is already present, stop copying do not update the children
      delete[] newChildren;
      return;
    }
  }

  newChildren[this->childCount] = child;
  child->parent = this;

  if (this->children != nullptr)
    delete[] this->children;

  this->children = newChildren;
  this->childCount = newChildCount;
}

Thing* Thing::RemoveChild(Thing* child) {
  unsigned char newChildCount = this->childCount - 1;
  Thing** newChildren = new Thing*[newChildCount];

  unsigned char newChildIx = 0;
  for (unsigned char childIx = 0; childIx < this->childCount; childIx++) {
    if (this->children[childIx] != child) {
      if (newChildIx == newChildCount) {  // We did not find the child
        // stop copying and return nothing
        delete[] newChildren;
        return nullptr;
      } else
        newChildren[newChildIx++] = this->children[childIx];
    }
  }

  child->parent = nullptr;

  delete[] this->children;
  this->children = newChildren;
  this->childCount = newChildCount;

  return child;
}

Thing* Thing::GetChild(unsigned char id, bool recursive) {
  for (unsigned char childIx = 0; childIx < this->childCount; childIx++) {
    Thing* child = this->children[childIx];
    if (child == nullptr)
      continue;
    if (child->id == id)
      return child;

    if (recursive) {
      Thing* foundChild = child->GetChild(id, recursive);
      if (foundChild != nullptr)
        return foundChild;
    }
  }
  return nullptr;
}

Thing* Thing::GetChildByIndex(unsigned char ix) {
  return this->children[ix];
}

void Thing::SetModel(const char* url) {
  this->modelUrl = url;
}

#pragma region Update

unsigned long Thing::GetTimeMs() {
#if defined(ARDUINO)
  return millis();
#else
  auto now = std::chrono::steady_clock::now();
  auto ms = std::chrono::duration_cast<std::chrono::milliseconds>(
      now.time_since_epoch());
  return static_cast<unsigned long>(ms.count());
#endif
}

void Thing::Update(bool recursive) {
  Update(GetTimeMs(), recursive);
}

void Thing::Update(unsigned long currentTimeMs, bool recursive) {
  // if (this->positionUpdated || this->orientationUpdated)
  //   OnPoseChanged callback
  this->positionUpdated = false;
  this->orientationUpdated = false;
  // this->linearVelocityUpdated = false;
  // this->angularVelocityUpdated = false;
  this->hierarchyChanged = false;
  this->nameChanged = false;

  if (recursive) {
    for (unsigned char childIx = 0; childIx < this->childCount; childIx++) {
      Thing* child = this->children[childIx];
      if (child == nullptr)
        continue;
      child->Update(currentTimeMs, recursive);
    }
  }
}

void Thing::UpdateThings(unsigned long currentTimeMs) {
  IsolatedParticipant::Isolated()->Update(currentTimeMs);
}

#pragma endregion Update

int Thing::GenerateBinary(char* buffer, unsigned char* ix) {
  (void)buffer;
  (void)ix;
  return 0;
}
void Thing::ProcessBinary(char* bytes) {
  (void)bytes;
};

void Thing::SetPosition(Spherical position) {
  this->position = position;
  this->positionUpdated = true;
}
Spherical Thing::GetPosition() {
  return this->position;
}

void Thing::SetOrientation(SwingTwist orientation) {
  this->orientation = orientation;
  this->orientationUpdated = true;
}

SwingTwist Thing::GetOrientation() {
  return this->orientation;
}

void Thing::SetLinearVelocity(Spherical linearVelocity) {
  if (this->linearVelocity.distance != linearVelocity.distance) {
    this->linearVelocity = linearVelocity;
    this->linearVelocityUpdated = true;
  }
}

Spherical Thing::GetLinearVelocity() {
  return this->linearVelocity;
}

void Thing::SetAngularVelocity(Spherical angularVelocity) {
  if (this->angularVelocity.distance != angularVelocity.distance) {
    this->angularVelocity = angularVelocity;
    this->angularVelocityUpdated = true;
  }
}

Spherical Thing::GetAngularVelocity() {
  return this->angularVelocity;
}

}  // namespace RoboidControl